//--------------------------------------------------------------------------------------
// File: Tutorial06.fx
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//--------------------------------------------------------------------------------------


//--------------------------------------------------------------------------------------
// Constant Buffer Variables
//--------------------------------------------------------------------------------------

cbuffer cbNeverChanges
    {
        matrix View;
		float4 vViewWorld;
    };
    
    cbuffer cbChangeOnResize
    {
        matrix Projection;
    };
    
    cbuffer cbChangesEveryFrame
    {
        matrix World;
		float4 vLightDir[2];
		float4 vLightColor[2];
    };

Texture2D txDecal;
SamplerState samLinear
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
};

RasterizerState noCull {
	CullMode = None;
};

//--------------------------------------------------------------------------------------
struct VS_INPUT
{
    float4 Pos  : POSITION;
    float2 Tex  : TEXCOORD0;
    float3 Norm : NORMAL;
};

struct PS_INPUT
{
    float4 Pos : SV_POSITION;
    float2 Tex : TEXCOORD0;
    float3 Norm : TEXCOORD1;
};


//--------------------------------------------------------------------------------------
// Vertex Shader
//--------------------------------------------------------------------------------------
PS_INPUT VS( VS_INPUT input )
{
    PS_INPUT output = (PS_INPUT)0;
    output.Pos  = mul( input.Pos,  World );
    output.Pos  = mul( output.Pos, View );
    output.Pos  = mul( output.Pos, Projection );
    output.Norm = mul( input.Norm, World ); // only valid, if World is rigid!!!!
    output.Tex = input.Tex;
    
    return output;
}

//--------------------------------------------------------------------------------------
// Pixel Shader
//--------------------------------------------------------------------------------------
float4 PS( PS_INPUT input) : SV_Target
{
    float4 finalColor = 0;
    
    float4 decalColor = txDecal.Sample( samLinear, input.Tex );
    float3 N = normalize(input.Norm.xyz);
    float3 V = normalize(vViewWorld.xyz);
    
    //do NdotL lighting for 2 lights
    for(int i=0; i<2; i++)
    {
		float3 H = normalize(vLightDir[i].xyz + V);
        finalColor += saturate( dot( (float3)vLightDir[i],input.Norm) * vLightColor[i] ) * decalColor;
        finalColor += 0.3*pow(saturate( dot( H,N)),1000) * vLightColor[i];
    }
    finalColor.a = 1;
    return finalColor;
}

//--------------------------------------------------------------------------------------
technique10 Render
{
    pass P0
    {
		SetRasterizerState(noCull);
        SetVertexShader( CompileShader( vs_4_0, VS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS() ) );
    }
}
